
#include <sys/file.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <ods2.h>

#include <err.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <mntopts.h>

struct mntopt mopts[] = {
	MOPT_STDOPTS,
	{ "lower", 0, ODS2MNT_LOWERCASE, 1 },
        { "nodirext", 0, ODS2MNT_NODIREXT, 1 },
        { "fastnodirext", 0, ODS2MNT_FAST_NODIREXT, 1 },
	{ NULL }
};

void	usage(void);

static char _revision_[] = "$Revision: 1.2 $";

int
main(int argc, char **argv)
{
	struct ods2_args args;
	int ch, mntflags, opts;
	char *dev, *dir;
	struct vfsconf vfc;
	int error, verbose;

	mntflags = opts = verbose = 0;

	memset(&args, 0, sizeof args);

	while ((ch = getopt(argc, argv, "fdlo:")) != -1)
		switch (ch) {
		case 'l':
			opts |= ODS2MNT_LOWERCASE;
			break;
                case 'd':
                        opts |= ODS2MNT_NODIREXT;
			break;
                case 'f':
                        opts |= ODS2MNT_FAST_NODIREXT;
			break;
		case 'o':
			getmntopts(optarg, mopts, &mntflags, &opts);
			break;
		case '?':
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 2)
		usage();

	dev = argv[0];
	dir = argv[1];

	/*
	 * ODS2 is not writeable
	 */

	mntflags |= MNT_RDONLY;

	args.export.ex_flags = MNT_EXRDONLY;
	args.fspec = dev;
	args.export.ex_root = -2;
	args.flags = opts;

	error = getvfsbyname("ods2", &vfc);
	if (error && vfsisloadable("ods2")) {
		if (vfsload("ods2"))
			err(EX_OSERR, "vfsload(ods2)");
		endvfsent();	/* flush cache */
		error = getvfsbyname("ods2", &vfc);
	}

	if (error)
		errx(1, "ods2 filesystem is not available");

	if (mount(vfc.vfc_name, dir, mntflags, &args) < 0)
		err(1, NULL);
	exit(0);
}

void
usage(void)
{
	(void)fprintf(stderr,
		"usage: mount_ods2 [-l] [-o options] special node\n");
	exit(EX_USAGE);
}
