
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/vnode.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/fcntl.h>
#include <sys/conf.h>

#if __FreeBSD__==3
#include <miscfs/specfs/specdev.h>
#include <sys/disklabel.h>
#else if __FreeBSD__==4
#include <sys/disk.h>
#include <sys/diskslice.h>
#endif

#include "ods2.h"

static char _revision_[] = "$Version$";

extern vop_t **ods2_vnodeop_p;

#define DEBUG_VGET 0
#define DEBUG_UNMOUNT 0

MALLOC_DEFINE(M_ODS2FSMNT, "ODS2FS mount", "ODS2FS mount structure");
MALLOC_DEFINE(M_ODS2FSFH, "ODS2FS file header", "ODS2FS file header");
MALLOC_DEFINE(M_ODS2FSNODE, "ODS2FS node", "ODS2FS inode structure");

static int	ods2_checksum __P((unsigned short *data, unsigned offset));
static int	ods2_homeblock_search __P((struct vnode *devvp,
                                           struct ods2_geometry *geometry,
                                           struct home_block_t *homeblock));
static int	ods2_perform_mount __P((struct vnode *devvp,
                                        struct mount *mp,
                                        struct proc *p,
                                        struct ods2_args *args));
static u_quad_t	ods2_file_size __P((struct file_header_t *header));

static int make_fid_directory __P((struct vnode *devvp,
                                   struct mount *mp,
                                   int fid_directory_fid,
                                   struct ods2_node *indexf_ip,
                                   struct vnode **vp));
static int ods2_check_header __P((char *buffer));

static u_quad_t ods2_file_size(header)
    struct file_header_t *header;
{
    u_quad_t size = 0;

    struct rms_attrib_t *rms = &header->fh2_w_recattr;
    size = INVERT(&rms->fat_l_efblk);
    size--;

    /*
     * Directories have a "fake" block on the start to hold "." and ".."
     */

    if (header->fh2_l_filechar & FCH_M_DIRECTORY)
        size++;

    size *= ODS2_BLOCK_SIZE;
    size += rms->fat_w_ffbyte;

    return size;
}

static int
ods2_checksum(data, offset)
    unsigned short *data;
    unsigned offset;
{
    unsigned int sum;
    unsigned int i;
    unsigned int s;

    sum = 0;
    s = offset >> 1;

    for (i=0; i < s; i++)
        sum += data[i];

    sum &= 0x0000FFFF;

    if (sum != data[s])
        return(ENOENT);

    return 0;
}

static int
ods2_homeblock_search(devvp, geometry, homeblock)
    struct vnode *devvp;
    struct ods2_geometry *geometry;
    struct home_block_t *homeblock;
{
    /*
     * XXX: To be correct, we need to perform the VMS home block search.
     *      Having coded this before, I'm not inclined to do it again.  Maybe
     *      someday I steal the code.
     */

    int error;
    unsigned lbn;
    struct home_block_t *hbp;
    unsigned char block[ODS2_BLOCK_SIZE];

    for (lbn=1; lbn < 30; lbn++)
    {
        error = ods2_bread(devvp, geometry, lbn, NOCRED, block);
        if (error)
            return error;

        error = ods2_checksum((unsigned short *) block, HM2_C_CHECKSUM1);
        if (error)
            continue;

        error = ods2_checksum((unsigned short *) block, HM2_C_CHECKSUM2);
        if (error)
            continue;

        /*
         * Okay, the checksums look okay, does the rest of it?
         */

        hbp = (struct home_block_t *) block;
        if (bcmp(hbp->hm2_t_format, DECFILE11B, sizeof(hbp->hm2_t_format)) != 0)
            continue;

        if (hbp->hm2_w_struclev.fh2_b_struclev_version != 1 ||
            hbp->hm2_w_struclev.fh2_b_struclev_level != 2)
            continue;

        /*
         * Looks like a good homeblock.  Let's use it!
         */

        bcopy(hbp, homeblock, sizeof(struct home_block_t));

        return 0;
    }

    return EINVAL;
}

int make_fid_directory(devvp, mp, fid, indexf_ip, vp)
    register struct vnode *devvp;
    struct mount *mp;
    int fid;
    struct ods2_node *indexf_ip;
    struct vnode **vp;
{
    int error;
    struct ods2_node *ip;
    struct fi2_t *fi2;
    dev_t dev = devvp->v_rdev;

if (0)
printf("make_fid_directory[starting]\n");

    MALLOC(ip,
           struct ods2_node *,
           sizeof(struct ods2_node),
           M_ODS2FSNODE,
           M_WAITOK);

    bzero((caddr_t) ip, sizeof(struct ods2_node));

    /*
     * Then get a vnode for it
     */

    if ((error = getnewvnode(VT_NON, mp, ods2_vnodeop_p, vp)) != 0)
    {
if (0)
printf("make_fid_directory[\tgetnewvnode failed=%d]\n", error);

        return error;
    }
    
    lockinit(&ip->i_lock, PINOD, "ods2node", 0, 0);
    
    (*vp)->v_data = ip;
    ip->i_vnode = *vp;
    ip->i_dev = dev;
    ip->i_devvp = devvp;
    ip->mount = mp;
    ip->i_num = fid;
    ip->i_size = 512;

    MALLOC(ip->file_header,
           struct file_header_t *,
           FH2_S_LENGTH,
           M_ODS2FSFH,
           M_WAITOK);

    bcopy(indexf_ip->file_header, ip->file_header, FH2_S_LENGTH);

    fi2 = FI2_AREA(ip->file_header);
    fi2->fi2_t_filename[0] = '#';

    /*
     * Don't bother to re-checksum it
     *
     * Put it in the hash table.  We'll never ask for it again, but it makes
     * unmounting easier if its in the hash table.  vput will want to reclaim
     * the node which looks in the hash table.  If it wasn't there to start
     * with, we crash - or we have to check for it.  This is simpler.
     */

    ods2_ihashins(ip);

    VREF(devvp);

    (*vp)->v_type = VDIR;

if (0)
printf("make_fid_directory[\tending]\n");

    return 0;
}

static int
ods2_perform_mount(devvp, mp, p, args)
    register struct vnode *devvp;
    struct mount *mp;
    struct proc *p;
    struct ods2_args *args;
{
    int error;
    unsigned v;
    unsigned i;
    unsigned lbn;
    dev_t dev = devvp->v_rdev;

    struct vnode *bitmap;
    struct ods2_node *ip;
    struct ods2_mount *imp;
    struct home_block_t homeblock;
    struct scb_t *scbp;
    struct ods2_geometry geometry;

    unsigned char block[ODS2_BLOCK_SIZE];

    /*
     * Disallow multiple mounts of the same device.
     * Disallow mounting of a device that is currently in use
     * (except for root, which might share swap device for miniroot).
     * Flush out any old buffers remaining from a previous use.
     */

    if ((error = vfs_mountedon(devvp)))
        return error;
    if (vcount(devvp) > 1 && devvp != rootvp)
        return EBUSY;
    if ((error = vinvalbuf(devvp, V_SAVE, p->p_ucred, p, 0, 0)))
        return (error);

    vn_lock(devvp, LK_EXCLUSIVE | LK_RETRY, p);

    if ((error = VOP_OPEN(devvp, FREAD, FSCRED, p)))
    {
        VOP_UNLOCK(devvp, 0, p);
        return error;
    }

    VOP_UNLOCK(devvp, 0, p);

    /*
     * Determine the sector size of the target disk.  This file system
     * assumes 512 bytes/sector, but that might not be the case.  We
     * get the sector size so we can "transparently" read the blocks
     * in the ods2_bread routine.
     */

    {
#if __FreeBSD__==3

        /*
         * FreeBSD version 3 requires an ioctl to get the sector size.  Its
         * stored in the disklabel.
         */

        struct partinfo dpart;

        error = VOP_IOCTL(devvp, DIOCGPART, (caddr_t)&dpart, FREAD, NOCRED, p);
        if (error != 0)
            return error;

        geometry.sector_size = dpart.disklab->d_secsize;

        if (0)
            printf("ods2[secsize=%d]\n", geometry.sector_size);

#else if __FreeBSD__==4

        /*
         * FreeBSD version 4 already has this information, we merely need to
         * access it.
         */

        struct disk *disk;
        struct diskslices *diskslices;

        disk = dev->si_disk;
        diskslices = disk->d_slice;

        geometry.sector_size = diskslices->dss_secsize;

        if (0)
            printf("ods2[secsize=%d, secshift=%d, secmult=%d]\n",
                   diskslices->dss_secsize,
                   diskslices->dss_secshift,
                   diskslices->dss_secmult);

#endif
    }


    /*
     * Now, calculate the block_shift.  This is used to convert 512 byte/sector
     * blocks into device blocks.
     *
     * device_blocks = 512  => block_shift = 0 => divide by 1
     * device_blocks = 1024 => block_shift = 1 => divide by 2
     * device_blocks = 2048 => block_shift = 2 => divide by 4
     * device_blocks = 4096 => block_shift = 3 => divide by 8
     */

    geometry.block_shift = ffs(geometry.sector_size) - 1;

    for (i=0,v=DEV_BSIZE; v <= 65536; v <<= 1,i++)
        if (v == geometry.sector_size)
            break;

    geometry.block_mask = 0x7FFFFFFF >> (31 - i);

    /*
     * Perform a home block search.  This may be a full home block search
     * based on mount flags.
     */

    if (0)
        printf("ods2_perform_mount[performing homeblock search]\n");

    error = ods2_homeblock_search(devvp, &geometry, &homeblock);
    if (error)
    {
        (void)VOP_CLOSE(devvp, FREAD, NOCRED, p);        
        return error;
    }

    MALLOC(imp,
           struct ods2_mount *,
           sizeof(struct ods2_mount),
           M_ODS2FSMNT,
           M_WAITOK);

    bzero(imp, sizeof(struct ods2_mount));

    mp->mnt_data = (qaddr_t) imp;
    imp->ods2_vfs = mp;
    imp->ods2_dev = dev;
    imp->ods2_devvp = devvp;
    imp->flags = args->flags;

    bcopy(&geometry, &imp->geometry, sizeof(struct ods2_geometry));

    /*
     * Extract the location of INDEXF.SYS file header.  We want the VBN.
     * Note:  this may also be calculated, per Kirby's book, as:
     *        v * 4 + m + n
     *        v - cluster size of disk
     *        m - size of index file bitmap
     *            (maxfiles+4096-1) / 4096
     *        n - file number
     */

    imp->maxfiles = homeblock.hm2_l_maxfiles;
    imp->fid_directory_fid = homeblock.hm2_l_maxfiles + 1;
    imp->indexf_vbn = homeblock.hm2_w_ibmapvbn + homeblock.hm2_w_ibmapsize;

    /*
     * Read the indexf.sys file header and verify
     */

    if (0)
        printf("ods2_perform_mount[reading INDEXF.SYS file header]\n");

    lbn = homeblock.hm2_l_ibmaplbn + homeblock.hm2_w_ibmapsize;

    error = ods2_bread(imp->ods2_devvp, &imp->geometry, lbn, NOCRED, block);
    if (error)
    {
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return error;
    }

    error = ods2_check_header(block);
    if (error)
    {
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return error;
    }
        
    /*  
     * Now, we have to make the inode.  First allocate some memory for it.
     */

    MALLOC(ip,
           struct ods2_node *,
           sizeof(struct ods2_node),
           M_ODS2FSNODE,
           M_WAITOK);

    bzero((caddr_t)ip, sizeof(struct ods2_node));

    /*
     * Then get a vnode for it
     */

    if ((error = getnewvnode(VT_NON, mp, ods2_vnodeop_p, &imp->indexf)) != 0)
    {
        FREE(ip, M_ODS2FSNODE);
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return (error);
    }   
    
    lockinit(&ip->i_lock, PINOD, "ods2node", 0, 0);
    
    imp->indexf->v_data = ip;
    ip->i_vnode = imp->indexf;
    ip->i_dev = dev;
    ip->i_devvp = imp->ods2_devvp;
    ip->mount = mp;
    ip->i_num = 1;     /* This is just the FILE_NUMBER in the fid_t */

    MALLOC(ip->file_header,
           struct file_header_t *,
           FH2_S_LENGTH,
           M_ODS2FSFH,
           M_WAITOK);

    if (0)
        printf("ods2_perform_mount[saving INDEXF.SYS file header]\n");

    bcopy(block, ip->file_header, FH2_S_LENGTH);

    /*
     * Put it in the hash table.  We'll never ask for it again, but it makes
     * unmounting easier if its in the hash table.  vput will want to reclaim
     * the node which looks in the hash table.  If it wasn't there to start
     * with, we crash - or we have to check for it.  This is simpler.
     */

    ods2_ihashins(ip);

    /*
     * Reference the device for the INDEXF.SYS file
     */

    VREF(devvp);

    /*
     * Access the MFD
     */

    error = ods2_vget(mp, (ino_t) 4, &imp->mfd);
    if (error)
    {
        vput(imp->indexf);
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return (error);
    }

    VOP_UNLOCK(imp->mfd, 0, p);

    /*
     * Access the SCB to get the total blocks.  Because INDEXF.SYS is
     * contiguous for the first 16 files, we can skip the VBN->LBN mapping
     * and just read the next header.  BITMAP.SYS is file number 2.
     */

    /*
     * Validate the SCB.  First, get BITMAP.SYS
     */

    if (0)
        printf("ods2_perform_mount[getting BITMAP.SYS]\n");

    error = ods2_vget(mp, (ino_t) 2, &bitmap);
    if (error)
    {
        vput(imp->mfd);
        vput(imp->indexf);
        FREE(ip, M_ODS2FSNODE);
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return (error);
    }

    /*
     * ... Now, read VBN=1 in BITMAP.SYS
     */

    if (0)
        printf("ods2_perform_mount[reading the SCB]\n");

    error = VOP_BMAP(bitmap, 1, NULL, &lbn, NULL, NULL);
    if (error)
    {
        vput(bitmap);
        vput(imp->mfd);
        vput(imp->indexf);

        FREE(ip, M_ODS2FSNODE);
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return (error);
    }

    error = ods2_bread(imp->ods2_devvp, &imp->geometry, lbn, NOCRED, block);
    if (error)
    {
        vput(bitmap);
        vput(imp->mfd);
        vput(imp->indexf);

        FREE(ip, M_ODS2FSNODE);
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return (error);
    }

    scbp = (struct scb_t *) block;

    imp->total_blocks = scbp->scb_l_volsize;

    vput(bitmap);

    /*
     * XXX: Need the FID directory.
     *
     * The FID directory is a special directory that doesn't actually exist
     * on ODS-2 volumes.  It is the "#" directory that is an invalid VMS
     * file name.  Its purpose is to allow access to files "by fid" under
     * Unix.  The user of the file system opens #/N where N is the fid of
     * the file to open/access.  The ods2_lookup routine knows to look for
     * lookups like this and will attempt to simply do ods2_vget using the
     * the N for the inode number.  The "#" directory lives in the root
     * and is visible when listing entries in the root.  As such, we must have
     * a valid vnode for it.  That is what we do here.
     */

    error = make_fid_directory(devvp,
                               mp,
                               imp->fid_directory_fid,
                               ip,
                               &imp->fid_directory);
    if (error)
    {
        vput(bitmap);
        vput(imp->mfd);
        vput(imp->indexf);

        FREE(ip, M_ODS2FSNODE);
        FREE(imp, M_ODS2FSMNT);
        (void) VOP_CLOSE(devvp, FREAD, NOCRED, p);
        return (error);
    }

/* XXX: DON'T NEED!    VREF(imp->fid_directory);  */

    /*
     * Link up all the structures
     */

    mp->mnt_stat.f_fsid.val[0] = (long)dev;
    mp->mnt_stat.f_fsid.val[1] = mp->mnt_vfc->vfc_typenum;
    mp->mnt_maxsymlinklen = 0;
    mp->mnt_flag |= MNT_LOCAL;

/*    devvp->v_specmountpoint = mp; */

    if (0)
        printf("ods2_perform_mount[done]\n");

    VOP_UNLOCK(imp->fid_directory, 0, p);
    VOP_UNLOCK(imp->indexf, 0, p);

#if __FreeBSD__==3

    (void)VOP_CLOSE(devvp, FREAD, NOCRED, p);        

#endif

    return 0;
}

/*
 * Mount ODS-2 layer
 */
int
ods2_mount(mp, path, data, ndp, p)
    struct mount *mp;
    char *path;
    caddr_t data;
    struct nameidata *ndp;
    struct proc *p;
{
    int error;
    mode_t accessmode;
    size_t size;
    struct vnode *devvp;
    struct ods2_args args;
    struct ods2_mount *imp;

    /*
     * Update is a no-op, must be read-only
     */

    if (mp->mnt_flag & MNT_UPDATE)
        return (EOPNOTSUPP);

    if ((mp->mnt_flag & MNT_RDONLY) == 0)
        return (EROFS);

    /*
     * Get argument
     */

    error = copyin(data, (caddr_t)&args, sizeof(struct ods2_args));
    if (error)
        return (error);

    /*
     * Look up the name and verify that it refers to a sensible block
     * device.
     */

    NDINIT(ndp, LOOKUP, FOLLOW, UIO_USERSPACE, args.fspec, p);
    if ((error = namei(ndp)))
        return (error);

    devvp = ndp->ni_vp;

#if __FreeBSD__==3

    if (devvp->v_type != VBLK) {
        vrele(devvp);
        return ENOTBLK;
    }

    if (major(devvp->v_rdev) >= nblkdev ||
        bdevsw[major(devvp->v_rdev)] == NULL)
    {
        vrele(devvp);
        return ENXIO;
    }   

#else if __FreeBSD__==4

    if (!vn_isdisk(devvp, &error)) {
        vrele(devvp);
        return (error);
    }

#endif

    /*      
     * Verify that user has necessary permissions on the device,
     * or has superuser abilities
     */

    accessmode = VREAD;
    vn_lock(devvp, LK_EXCLUSIVE | LK_RETRY, p);
    error = VOP_ACCESS(devvp, accessmode, p->p_ucred, p);

#if __FreeBSD__==3

    if (error)
        error = suser(p->p_ucred, &p->p_acflag);

#else if __FreeBSD__==4

    if (error)
        error = suser(p);

#endif

    if (error) {
        vput(devvp);
        return (error);
    }

    VOP_UNLOCK(devvp, 0, p);

    /*
     * And finally do the actual mount
     */

    error = ods2_perform_mount(devvp, mp, p, &args);
    if (error)
    {
        vrele(devvp);
        return error;
    }

    imp = MOUNTTOODS2MOUNT(mp);
    (void) copyinstr(path, mp->mnt_stat.f_mntonname, MNAMELEN - 1, &size);
    bzero(mp->mnt_stat.f_mntonname + size, MNAMELEN - size);
    (void) copyinstr(args.fspec,
                     mp->mnt_stat.f_mntfromname,
                     MNAMELEN - 1,
                     &size);
    bzero(mp->mnt_stat.f_mntfromname + size, MNAMELEN - size);
    (void) ods2_statfs(mp, &mp->mnt_stat, p);

    return 0;
}

int
ods2_unmount(mp, mntflags, p)
	struct mount *mp;
	int mntflags;
	struct proc *p;
{
    int error;
    int flags = 0;
    struct ods2_mount *imp = MOUNTTOODS2MOUNT(mp);
    struct vnode *devvp = imp->ods2_devvp;

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[starting]\n");

    if (mntflags & MNT_FORCE)
        flags |= FORCECLOSE;

#if 0
	mntflushbuf(mp, 0);
	if (mntinvalbuf(mp, 1))
		return (EBUSY);
#endif

    if (imp->mfd->v_usecount > 2)
    {
        if (DEBUG_UNMOUNT)
            printf("ods2_unmount[\tbusy]\n");

	return (EBUSY);
    }

    if (DEBUG_UNMOUNT)
    {
        printf("ods2_unmount[mfd=%d]\n", imp->mfd->v_usecount);
        printf("ods2_unmount[fid=%d]\n", imp->fid_directory->v_usecount);
        printf("ods2_unmount[indexf=%d]\n", imp->indexf->v_usecount);
        printf("ods2_unmount[devvp=%d]\n", imp->ods2_devvp->v_usecount);
    }

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[releasing MFD...]\n");

    vrele(imp->mfd);

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[releasing FID...]\n");

    vrele(imp->fid_directory);

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[after vrele of FID, devvp=%d]\n",
               imp->ods2_devvp->v_usecount);

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[releasing INDEXF...]\n");

    vrele(imp->indexf);

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[after vrele, devvp=%d]\n",
               imp->ods2_devvp->v_usecount);

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[\tvflush'ing...]\n");

    error = vflush(mp, NULLVP, flags);
    if (error)
    {
        if (DEBUG_UNMOUNT)
            printf("ods2_unmount[\tvflush error]\n");

        return (error);
    }

#if __FreeBSD__ == 4

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[VOP_CLOSE on devvp]\n");

    (void)VOP_CLOSE(devvp, FREAD, NOCRED, p);

#endif

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[releasing devvp...]\n");

    vrele(devvp);

/* XXX: free indexf, fid, etc.? */

    /*
     * Finally, throw away the ods2_mount structure
     */

    FREE(mp->mnt_data, M_ODS2FSMNT);     /* XXX */
    mp->mnt_data = 0;

    if (DEBUG_UNMOUNT)
        printf("ods2_unmount[\tthat's all]\n");

    return 0;
}

int
ods2_root(mp, vpp)
	struct mount *mp;
	struct vnode **vpp;
{
    int error;
    struct vnode *vp;

if (0)
printf("ods2_root[starting]\n");

    error = VFS_VGET(mp, (ino_t) MFD_FILE_NUMBER, &vp);
    if (error)
    {
if (0)
printf("ods2_root[\terror=%d]\n", error);
        return error;
    }

    vp->v_flag |= VROOT;

    *vpp = vp;

if (0)
printf("ods2_root[\tgood]\n");

    return(0);
}

/*
 * We could actually do something with this, but since we're read-only and
 * this is my first attempt at a file system, we'll keep it simple!
 */

int
ods2_quotactl(mp, cmd, uid, arg, p)
	struct mount *mp;
	int cmd;
	uid_t uid;
	caddr_t arg;
	struct proc *p;
{
    return(EOPNOTSUPP);
}

/*
 * Because we are read-only, we'll fudge the free stuff.
 */

int
ods2_statfs(mp, sbp, p)
	struct mount *mp;
	struct statfs *sbp;
	struct proc *p;
{
    register struct ods2_mount *imp;

if (0)
printf( "ods2_statfs[starting]\n" );

    imp = MOUNTTOODS2MOUNT(mp);

    sbp->f_bsize = ODS2_BLOCK_SIZE;
    sbp->f_iosize = sbp->f_bsize;       /* XXX */
    sbp->f_blocks = imp->total_blocks;
    sbp->f_bfree = 0;                   /* total free blocks */
    sbp->f_bavail = 0;                  /* blocks free for non superuser */
    sbp->f_files =  0;                  /* total files */
    sbp->f_ffree = 0;                   /* free file nodes */
    if (sbp != &mp->mnt_stat) {
        sbp->f_type = mp->mnt_vfc->vfc_typenum;
        bcopy(mp->mnt_stat.f_mntonname, sbp->f_mntonname, MNAMELEN);
        bcopy(mp->mnt_stat.f_mntfromname, sbp->f_mntfromname, MNAMELEN);
    }
    return 0;
}

int ods2_check_header( char *buffer )
{
    unsigned offset;
    int file_number;
    int error;
    struct file_header_t *tfh2;

    error = ods2_checksum((unsigned short *) buffer, FH2_C_CHECKSUM);
    if (error)
    {
        if (0)
            printf("ods2_check_header[checksum failed]\n");

        return error;
    }

    tfh2 = (struct file_header_t *) buffer;

    if ( tfh2->fh2_b_idoffset < FH2_S_HIGHWATER_D2 )
    {
        if (0)
            printf("ods2_check_header[highwater vs. id offset failed]\n");
 
        return ENOENT;
    }

    if ( ! (tfh2->fh2_b_idoffset <= tfh2->fh2_b_mpoffset &&
            tfh2->fh2_b_mpoffset <= tfh2->fh2_b_acoffset &&
            tfh2->fh2_b_acoffset <= tfh2->fh2_b_rsoffset) )
    {
        if (0)
            printf("ods2_check_header[transitive offset check failed]\n");

        return ENOENT;
    }

    if ( tfh2->fh2_w_struclev.fh2_b_struclev_version != 1 ||
         tfh2->fh2_w_struclev.fh2_b_struclev_level != 2 )
    {
        if (0)
            printf("ods2_check_header[structure check failed]\n");

        return ENOENT;
    }

    offset = tfh2->fh2_b_acoffset - tfh2->fh2_b_mpoffset;
    if (tfh2->fh2_b_map_inuse > offset)
    {
        if (0)
           printf("ods2_check_header[map inuse vs. size of map area failed]\n");

        return ENOENT;
    }

    /* VALID FILE HEADER */

    file_number = FILE_NUMBER(tfh2->fh2_w_fid);
    if (file_number == 0 &&
        tfh2->fh2_w_checksum == 0 &&
        (tfh2->fh2_l_filechar & FCH_M_MARKDEL) != 0)
    {
        /* But this file was deleted :( */

        if (0)
            printf("ods2_check_header[deleted file]\n");

        return ENOENT;
    }

    return 0;
}

int ods2_read_file_header( struct mount *mp,
                           ino_t file_number,
                           struct file_header_t **fh2 )
{
    struct ods2_mount *imp;
    struct vnode *indexf_vp;
    dev_t dev;
    int error;
    int lbn;
    int vbn;
    unsigned char block[ODS2_BLOCK_SIZE];

    /*
     * We could VOP_READ using VBN on INDEXF.SYS, but we want this to be
     * a "lower" level utility routine.  Therefore, we'll do some of that
     * work here.
     */

    imp = MOUNTTOODS2MOUNT(mp);
    indexf_vp = imp->indexf;
    dev = imp->ods2_dev;

    vbn = imp->indexf_vbn + file_number - 1;

    if (DEBUG_VGET)
        printf("ods2_vget[\tgetting header: vbn=%d...]\n", vbn);

    error = VOP_BMAP(indexf_vp, vbn, NULL, &lbn, NULL, NULL);

    if (error)
    {
        if (DEBUG_VGET)
            printf("ods2_vget[\terror getting header lbn.  errno=%d]\n", error);

        return(error);
    }

    if (DEBUG_VGET)
        printf("ods2_vget[\theader at lbn=%d...]\n", lbn);

    error = ods2_bread(imp->ods2_devvp, &imp->geometry, lbn, NOCRED, block);
    if (error)
    {
        if (DEBUG_VGET)
            printf("ods2_vget[\terror reading header lbn.  errno=%d]\n", error);

        return error;
    }

    /*
     * Checksum the file header.  It may fail because this file was never on
     * the surface to start with!
     */

    if (DEBUG_VGET)
        printf("ods2_vget[\tchecksuming header...]\n");

    error = ods2_check_header(block);
    if (error)
    {

        if (DEBUG_VGET)
            printf("ods2_vget[\tbad checksum on header...]\n");

        return error;
    }

    MALLOC(*fh2, struct file_header_t *, FH2_S_LENGTH, M_ODS2FSFH, M_WAITOK);

    bcopy(block, *fh2, FH2_S_LENGTH);

    return 0;
}

/*
 * Given an inode number (which is internally generated), return its vnode.
 * The vnode may be in our cache as an inode or we may have to make both.
 * If the inode is in the cache, it will (should) have a pointer back to its
 * vnode, so just return the vnode.  If we have to make both, make the vnode
 * (that's easy), but read the inode from disk.
 *
 * We can easily read the inode from disk because we have its file number and
 * the ods2_mount structure contains the LBN of the start of file headers in
 * INDEXF.SYS.
 *
 * Now, actually getting the correct LBN may be difficult because INDEXF.SYS
 * may have several mapping pointers that we have to trapse through.
 */

int
ods2_vget(mp, ino, vpp)
    struct mount *mp;
    ino_t ino;
    struct vnode **vpp;
{
    struct ods2_mount *imp;
    struct ods2_node *ip;
    struct vnode *vp;
    struct vnode *indexf_vp;
    struct file_header_t *fh2;
    dev_t dev;
    int error;

    imp = MOUNTTOODS2MOUNT(mp);
    indexf_vp = imp->indexf;
    dev = imp->ods2_dev;

    if (DEBUG_VGET)
        printf("ods2_vget[starting, ino=%d\n", ino);

    /*
     * Easy case is its in the cache ;)
     */

    if ((*vpp = ods2_ihashget(dev, ino)) != NULLVP)
    {
        if (DEBUG_VGET)
            printf("ods2_vget[\tin cache]\n");

        return (0);
    }

    if ( ino == imp->fid_directory_fid )
        panic("FID Directory should be in ods2 inode cache");


    error = ods2_read_file_header( mp, ino, &fh2 );

    if (error)
    {
        if (DEBUG_VGET)
            printf("ods2_vget[\terror getting header lbn.  errno=%d]\n", error);

        return(error);
    }

    /*
     * Okay, header checks out.  Make sure its not a continuation header...
     */

    if (fh2->fh2_w_seg_num != 0)
    {
        FREE( fh2, M_ODS2FSFH );
        return ENOENT;
    }

    /*
     * Okay, now we have to make both the vnode and inode.
     */

    if (DEBUG_VGET)
    {
        printf("ods2_vget[\t%d not in cache, making]\n", ino);
        printf("ods2_vget[\tallocating memory of ods2_node]\n");
    }

    MALLOC(ip,
           struct ods2_node *,
           sizeof(struct ods2_node),
           M_ODS2FSNODE,
           M_WAITOK);

    bzero((caddr_t)ip, sizeof(struct ods2_node));

    /*
     * The vnode is easy.
     * We pass in our vnodeop_p so that the new vnode will be initialized with
     * our operations.
     */

/*    if ((error = getnewvnode(VT_ODS2FS, mp, ods2_vnodeop_p, &vp)) != 0) */

    if (DEBUG_VGET)
        printf("ods2_vget[\tgetnewvnode...]\n");

    if ((error = getnewvnode(VT_NON, mp, ods2_vnodeop_p, &vp)) != 0)
    {
        if (DEBUG_VGET)
            printf("ods2_vget[\tcan't get new vnode: %d]\n", error);

        FREE( fh2, M_ODS2FSFH );
        FREE(ip, M_ODS2FSNODE);
        *vpp = NULLVP;
        return (error);
    }

    if (DEBUG_VGET)
        printf("ods2_vget[\tlockinit...]\n");

    lockinit(&ip->i_lock, PINOD, "ods2node", 0, 0);

    vp->v_data = ip;
    ip->i_vnode = vp;
    ip->i_dev = dev;
    ip->i_devvp = imp->ods2_devvp;
    ip->i_num = ino;     /* This is just the FILE_NUMBER in the fid_t */
    ip->mount = mp;

    if (DEBUG_VGET)
        printf("ods2_vget[\tinserting into hashtable...]\n");

    ods2_ihashins(ip);

    /*
     * Okay, we're now in the home stretch!
     */

    if (DEBUG_VGET)
        printf("ods2_vget[\tallocating memory for header]\n");

    ip->file_header = fh2;

    if (DEBUG_VGET)
        printf("ods2_vget[\tVREF dev]\n");

    VREF(imp->ods2_devvp);

    ip->i_size = ods2_file_size(ip->file_header);

    if (DEBUG_VGET)
        printf("ods2_vget[file size=%d]\n", (int) ip->i_size);

    if (ino == 4) /* MFD */
        vp->v_flag |= VROOT;

    if (0)
        printf("ods2_vget[fh2_l_filechar=%x]\n",
               ip->file_header->fh2_l_filechar);
    if (0)
        printf("ods2_vget[fch_m_directory=%x]\n", FCH_M_DIRECTORY);
    if (0)
        printf("ods2_vget[filechar & directory=%x]\n",
               (ip->file_header->fh2_l_filechar & FCH_M_DIRECTORY));
       
    if (ip->file_header->fh2_l_filechar & FCH_M_DIRECTORY)
        vp->v_type = VDIR;
    else
        vp->v_type = VREG;

    *vpp = vp;

    if (DEBUG_VGET)
        printf("ods2_vget[that's all folks]\n");

if (0)
{
    int lbn;

    VOP_BMAP(vp, 1, NULL, &lbn, NULL, NULL);
    printf("%d first LBN: %d\n", ino, lbn);
}

    return(0);
}

struct ods2_fh
{
    unsigned short length;
    unsigned short pad;
    struct fid_t   fid;
};

int
ods2_vptofh(vp, fhp)
	struct vnode *vp;
	struct fid *fhp;
{
    struct ods2_node *ip = VTOODS2(vp);
    struct ods2_fh   *fh = (struct ods2_fh *) fhp;

    fh->length = sizeof(struct ods2_fh);
    bcopy(&ip->file_header->fh2_w_fid, &fh->fid, sizeof(struct fid_t));

    return 0;
}

/*               ********** VERSION DEPENDENT SECTION **********             */
/*
 * ods2_sync      v3 ONLY
 * ods2_start     v3 ONLY
 * ods2_fhtovp    v3->v4 change
 * ods2_checkexp  v4 ONLY
 */

#if __FreeBSD__==3

/*
 * HA!  We're read-only!
 */

int
ods2_sync(mp, waitfor, cred, p)
	struct mount *mp;
	int waitfor;
	struct ucred *cred;
	struct proc *p;
{
    return (0);
}

/*
 * No other file system does anything here :)
 */

int
ods2_start(mp, flags, p)
	struct mount *mp;
	int flags;
	struct proc *p;
{
    return (0);
}

int
ods2_fhtovp(mp, fidp, nam, vpp, exflagsp, credanonp)
	struct mount *mp;
	struct fid *fidp;
	struct sockaddr *nam;
	struct vnode **vpp;
	int *exflagsp;
	struct ucred**credanonp;
{
    struct ods2_fh *ifhp = (struct ods2_fh *) fidp;
    register struct ods2_node *ip;
    register struct netcred *np;
    register struct ods2_mount *imp = MOUNTTOODS2MOUNT(mp);
    struct vnode *nvp;
    int error;

    /*
     * Okay to export?
     */

    np = vfs_export_lookup(mp, &imp->export, nam);
    if (np == NULL)
        return (EACCES);

    /*
     * Okay to export.  Go get the vnode for this file.  The file to get is
     * given by the FILE_NUMBER in the handle's fid.
     */

    if ((error = VFS_VGET(mp, FILE_NUMBER(ifhp->fid), &nvp)) != 0)
    {
        *vpp = NULLVP;
        return (error);
    }

    /*
     * Got the vnode, change to ods2_node and check the sequence numbers.
     */

    ip = VTOODS2(nvp);
    if (ip->file_header->fh2_w_fid.fid_w_seq != ifhp->fid.fid_w_seq)
    {
        vput(nvp);
        *vpp = NULLVP;
        return (ESTALE);
    }

    /*
     * Ready to rumble...
     */

    *vpp = nvp;
    *exflagsp = np->netc_exflags;
    *credanonp = &np->netc_anon;

    return (0);
}
#else if __FreeBSD__==4

int
ods2_checkexp(mp, nam, exflagsp, credanonp)
        struct mount *mp;
        struct sockaddr *nam;
        int *exflagsp;
        struct ucred **credanonp;
{
        register struct netcred *np;
        register struct ods2_mount *imp;

        imp = MOUNTTOODS2MOUNT(mp);

        /*
         * Get the export permission structure for this <mp, client> tuple.
         */

        np = vfs_export_lookup(mp, &imp->export, nam);
        if (np == NULL)
                return (EACCES);
  
        *exflagsp = np->netc_exflags;
        *credanonp = &np->netc_anon;
        return (0);
} 

int
ods2_fhtovp(mp, fidp, vpp)
	struct mount *mp;
	struct fid *fidp;
	struct vnode **vpp;
{
    struct ods2_fh *ifhp = (struct ods2_fh *) fidp;
    register struct ods2_node *ip;
    struct vnode *nvp;
    int error;

    if ((error = VFS_VGET(mp, FILE_NUMBER(ifhp->fid), &nvp)) != 0)
    {
        *vpp = NULLVP;
        return (error);
    }

    /*
     * Got the vnode, change to ods2_node and check the sequence numbers.
     */

    ip = VTOODS2(nvp);
    if (ip->file_header->fh2_w_fid.fid_w_seq != ifhp->fid.fid_w_seq)
    {
        vput(nvp);
        *vpp = NULLVP;
        return (ESTALE);
    }

    /*
     * Ready to rumble...
     */

    *vpp = nvp;

    return (0);
}

#endif

#if __FreeBSD__==3

static struct vfsops ods2_vfsops = {
	ods2_mount,
	ods2_start,
	ods2_unmount,
	ods2_root,
	ods2_quotactl,
	ods2_statfs,
	ods2_sync,
	ods2_vget,
	ods2_fhtovp,
	ods2_vptofh,
	ods2_init
};

VFS_SET(ods2_vfsops, ods2, VFCF_READONLY);

#else if __FreeBSD__==4

static struct vfsops ods2_vfsops = {
	ods2_mount,
        vfs_stdstart,
	ods2_unmount,
	ods2_root,
        ods2_quotactl,
	ods2_statfs,
        vfs_stdsync,
	ods2_vget,
	ods2_fhtovp,
	ods2_checkexp,
	ods2_vptofh,
	ods2_init,
        vfs_stduninit,
        vfs_stdextattrctl
};

VFS_SET(ods2_vfsops, ods2, VFCF_READONLY);

#endif
